package OSM::Filter;

=encoding utf8

=head1 NAME

OSM::Filter -- Filter OpenStreetMap Data

=head1 SYNOPSIS

 #!/usr/bin/perl
 use strict;
 use warnings;
 use OSM::Filter;

 my $filter=OSM::Filter->new(infile=>'test.osm', outfile=>'result.osm');
 $filter->filter_add('way', sub{ my $way=shift; return 1 if($way->has_tag('highway')); return 0});
 $filter->run();

 # or

 my $filter=OSM::Filter->new(infile=>'test.osm');
 $filter->filter_add('way', sub{ my $way=shift; return 1 if($way->has_tag('highway')); return 0});
 my $osm=$filter->run(); # <= returns an OSM::Tree

=head1 DESCRIPTION

Filter OpenStreetMap XML file

=head1 AUTHOR

ToPeG

=head1 METHODS

=over 4

=cut

use strict;
use warnings;
use XML::Parser;
use IO::Scalar;
use OSM::Tree;
use Carp;

our $VERSION=0.1;

=item new( (optionname => option)+ )

create a new Filter-Object

returns an OSM::Filter Object

Options:

- infile

=over 4

an input file

=back

- outfile

=over 4

the output file

=back

- fh

=over 4

an input filehandle

=back

- data

=over 4

an OSM-XML String

=back

- filter

=over 4

an array of filter:

=over 4

filter:

 ['<key>',sub{<code>},<options>]

=over 4

key = [node, way, relation]

=back

=back

=back

=cut
sub new
{
  my $class=shift;
  $class=ref($class) if(ref($class));

  my $conf={};

  if(ref($_[0]) eq 'HASH')
  { $conf={ %{$_[0]} }; }
  else
  {
    pop(@_) if(@_%2);
    $conf={@_};
  }

  if($conf->{infile} && -f $conf->{infile})
  {
    unless(open($conf->{fh}, '<', $conf->{infile}))
    {
      carp(qq'ERROR open "$conf->{file}" ($!)\n');
      return undef;
    }
  }
  elsif($conf->{data})
  {
    unless( ref($conf->{data}) eq 'SCALAR' )
    {
      my $data=$conf->{data};
      $conf->{data}=\$data;
    }
    $conf->{fh} = new IO::Scalar $conf->{data};
  }
  elsif(!$conf->{fh})
  {
    carp(qq'No input given!');
    return undef;
  }

  my $self={};
  $self->{file} = $conf->{outfile};
  $self->{fh}   = $conf->{fh};

  bless($self,$class);

  if($conf->{filter})
  {
    my @filter=ref($conf->{filter}) eq 'ARRAY'?@{$conf->{filter}}:($conf->{filter});
    $self->filter_add($_) for(@filter);
  }

  return $self;
}

=item file_out(<Filename>)

get/set output file

=cut
sub file_out
{
  my $self=shift;
  $self->{file}=shift if(@_);
  return $self->{file};
}

=item file_in(<Filename>)

set input file

=cut

sub file_in
{
  my $self=shift;
  my $file=shift || return 0;
  my $fh;
  unless(open($fh, '<', $file))
  {
    carp(qq'ERROR open "$file" ($!)\n');
    return 0;
  }
  $self->{fh}=$fh;
  return 1;
}

=item filter_add(<Key>, <Code>, <Options>)

add an filter

=cut
sub filter_add
{
  my $self=shift;
  my $filter=shift;
  my ($key,$code,$opts);
  if(ref($filter) eq 'ARRAY')
  {
    $key=$filter->[0];
    $code=$filter->[1];
    $opts=$filter->[2];
  }
  elsif(ref($filter) eq 'HASH')
  {
    $key=$filter->{key};
    $code=$filter->{code};
    $opts=$filter->{opts};
  }
  elsif(@_)
  {
    $key=$filter;
    $code=shift;
    $opts=shift;
  }
  else
  { return 0; }

  return 0 unless($key && grep{$key eq $_}qw(node way relation));
  return 0 unless($code && ref($code) eq 'CODE');

  push(@{$self->{filter}->{$key}},[$code,$opts]);

  return 1;
}

=item run()

run filter

return OSM::Tree if no outputfile is set
else return 1 on success

=cut
sub run
{
  my $self=shift;

  my $tree={};
  my $tree_out={};
  my $inside;
  my $key;
  my $id;

  my $parser = XML::Parser->new(Handlers => {
    Start => sub{
      my $parser=shift;
      my $name=shift;
      my $opts={@_};

      if($name eq 'osm')
      {
        while(my ($k,$v) = each(%$opts))
        {
          $tree->{$k}=$v;
          $tree_out->{$k}=$v;
        }
      }
      elsif($name eq 'bounds')
      {
        push(@{$tree->{bounds}},$opts);
        #push(@{$tree_out->{bounds}},$opts);
      }
      elsif($name eq 'bound')
      {
        my @box=split(",",$opts->{box});
        delete($opts->{box});
        for my $k (qw(minlat minlon maxlat maxlon))
        { $opts->{$k}=shift(@box); }

        push(@{$tree->{bounds}},$opts);
      }
      elsif($name eq 'node' || $name eq 'way' || $name eq 'relation')
      {
        $id=$opts->{id};
        $key=$name;
        $tree->{$name}->{$id}=$opts;
        $inside=$opts;
      }
      elsif($inside)
      {
        if($name eq 'tag')
        {
          my $k=$opts->{k};
          my $v=$opts->{v};
          $inside->{tag}->{$k}=$v;
        }
        elsif($name eq 'nd')
        { push(@{$inside->{nd}},$opts->{ref}); }
        elsif($name eq 'member')
        {
          my $id =$opts->{ref};
          delete($opts->{ref});
          $inside->{member}->{$id}=$opts;
        }
        else
        { die("UNKOWN SUBTYPE $name\n"); }
      }
      else
      { die("UNKOWN TYPE $name\n"); }
    },
    End=>sub{
      my $parser=shift;
      my $name=shift;
      if($key && $name eq $key)
      {
        if($self->{filter}->{$name} && ref($self->{filter}->{$name}) eq 'ARRAY')
        {
          my $obj;
          if($name eq 'node')
          { $obj=OSM::Tree::Node->_new($tree,$id); }
          elsif($name eq 'way')
          { $obj=OSM::Tree::Way->_new($tree,$id); }
          elsif($name eq 'relation')
          { $obj=OSM::Tree::Relation->_new($tree,$id); }
          else
          { die("UNKOWN TYPE $name\n"); }

          my $ok=1;
          for(@{$self->{filter}->{$name}})
          { $ok=0 unless($_->[0]->($obj,$_->[1])); }

          if($ok)
          {
            if($name eq 'node')
            { __insert_node($tree,$tree_out,$id); }
            elsif($name eq 'way')
            { __insert_way($tree,$tree_out,$id); }
            elsif($name eq 'relation')
            { __insert_relation($tree,$tree_out,$id); }
          }
        }
        $key=undef;
        $id=undef;
      }
    }});

  $parser->parse($self->{fh});

  my $osm=OSM::Tree->new();
  $osm->_set_tree($tree_out);

  return __save($self->{file},$osm) if($self->{file});
  return $osm;
}

########################################################################
# private

sub __insert_node
{
  my $tree=shift;
  my $tree_out=shift;
  my $id=shift;
  $tree_out->{node}->{$id}=$tree->{node}->{$id} if(exists($tree->{node}->{$id}));
}

sub __insert_way
{
  my $tree=shift;
  my $tree_out=shift;
  my $id=shift;
  if(exists $tree->{way}->{$id} )
  {
    my $data=$tree->{way}->{$id};
    $tree_out->{way}->{$id}=$data;
    __insert_node($tree,$tree_out,$_) for(@{$data->{nd}})
  }
}

sub __insert_relation
{
  my $tree=shift;
  my $tree_out=shift;
  my $id=shift;
  if(exists($tree->{relation}->{$id}))
  {
    my $data=$tree->{relation}->{$id};
    $tree_out->{relation}->{$id}=$data;
    for(@{$data->{member}})
    {
      my $mky=$_->{type};
      my $mid=$_->{ref};
      if($mky eq 'way')
      { __insert_way($tree,$tree_out,$mid); }
      elsif($mky eq 'relation')
      { __insert_relation($tree,$tree_out,$mid);  }
      else
      { __insert_node($tree,$tree_out,$mid) }
    }
  }
}

sub __save
{
  my $file=shift;
  my $osm=shift;
  open(my $fh,'>:encoding(UTF-8)',$file) or return 0;
  print $fh $osm->to_xml();
  close($fh) or return 0;
  return 1;
}
=back

=head1 AUTHOR

Tobias Grönhagen <osm@topeg.de>

=head1 VERSION

Version 0.1 11 Apr 2011

=head1 COPYRIGHT

Copyright (c) Tobias Grönhagen. All rights reserved.

=head1 LICENSE

This program is free software; you can redistribute it
and/or modify it under the same terms and conditions as
Perl itself.

This means that you can, at your option, redistribute it and/or modify it under
either the terms the GNU Public License (GPL) version 2 or later, or under the
Perl Artistic License.

See http://dev.perl.org/licenses/

=head1 DISCLAIMER

THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE.

Use of this software in any way or in any form, source or binary,
is not allowed in any country which prohibits disclaimers of any
implied warranties of merchantability or fitness for a particular
purpose or any disclaimers of a similar nature.

IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
SPECIAL, INCIDENTAL,  OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
USE OF THIS SOFTWARE AND ITS DOCUMENTATION (INCLUDING, BUT NOT
LIMITED TO, LOST PROFITS) EVEN IF I HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE

=cut

1;
